import * as cdk from 'aws-cdk-lib';
import * as lambda from 'aws-cdk-lib/aws-lambda';
import * as lambdaNodejs from 'aws-cdk-lib/aws-lambda-nodejs';
import * as apigateway from 'aws-cdk-lib/aws-apigateway';
import * as dynamodb from 'aws-cdk-lib/aws-dynamodb';
import * as path from 'path';
import * as cognito from 'aws-cdk-lib/aws-cognito';
import { Effect, PolicyStatement } from 'aws-cdk-lib/aws-iam';

export class BackendAwsStack extends cdk.Stack {
  constructor(scope: cdk.App, id: string, props?: cdk.StackProps) {
    super(scope, id, props);

    // Create Cognito User Pool
    const userPool = new cognito.UserPool(this, 'ContactsAppUserPool', {
      selfSignUpEnabled: true,
      signInAliases: { email: true },
      autoVerify: { email: true },
      userPoolName: 'ContactsAppUserPool',
    });

    const userPoolClient = new cognito.UserPoolClient(
      this,
      'ContactsAppUserPoolClient',
      {
        userPoolClientName: 'ContactAppUserPoolClient',
        userPool,
        authFlows: {
          userPassword: true,
        },
      }
    );

    // DynamoDB Table for Contacts
    const contactsTable = new dynamodb.Table(this, 'Contacts-app_Table', {
      partitionKey: { name: 'id', type: dynamodb.AttributeType.STRING },
      tableName: 'Contacts',
      removalPolicy: cdk.RemovalPolicy.DESTROY, // Remove table when stack is destroyed (for dev purposes)
    });

    // Define Lambda Functions

    //Create Contact
    const createContactFunction = this.createLambdaFunction(
      'CreateContactFunction',
      '../functions/create-contact/index.ts',
      contactsTable
    );

    createContactFunction.addToRolePolicy(
      new PolicyStatement({
        effect: Effect.ALLOW,
        resources: [contactsTable.tableArn],
        actions: [
          'dynamodb:PutItem',
          'dynamodb:GetItem',
          'dynamodb:UpdateItem',
        ],
      })
    );

    //Get Contacts
    const getContactFunction = this.getLambdaFunction(
      'GetContactFunction',
      '../functions/get-contact/index.ts',
      contactsTable
    );

    getContactFunction.addToRolePolicy(
      new PolicyStatement({
        effect: Effect.ALLOW,
        resources: [contactsTable.tableArn],
        actions: ['dynamodb:Scan', 'dynamodb:GetItem'],
      })
    );

    /*

    const updateContactFunction = this.createLambdaFunction(
      'UpdateContactFunction',
      '../functions/update-contact/index.ts',
      contactsTable
    );

    const deleteContactFunction = this.createLambdaFunction(
      'DeleteContactFunction',
      '../functions/delete-contact/index.ts',
      contactsTable
    ); */

    // API Gateway
    const api = new apigateway.RestApi(this, 'ContactsApi', {
      restApiName: 'Contacts-app_Service',
      description: 'API for managing contacts.',
    });

    // Cognito Authorizer to API
    const authorizer = new apigateway.CognitoUserPoolsAuthorizer(
      this,
      'ContactsAppAuthorizer',
      {
        cognitoUserPools: [userPool],
      }
    );

    // API Routes
    const contactsResource = api.root.addResource('contacts');
    //const contactIdResource = contactsResource.addResource('{contactId}');

    contactsResource.addMethod(
      'POST',
      new apigateway.LambdaIntegration(createContactFunction),
      {
        authorizer,
        authorizationType: apigateway.AuthorizationType.COGNITO,
      }
    );

    contactsResource.addMethod(
      'GET',
      new apigateway.LambdaIntegration(getContactFunction),
      {
        authorizer,
        authorizationType: apigateway.AuthorizationType.COGNITO,
        requestParameters: { 'method.request.path.contacts': false },
      }
    );

    /*     // Get Contact - GET /contacts/{id}
    const singleContactResource = contactsResource.addResource('{id}');
    singleContactResource.addMethod(
      'GET',
      new apigateway.LambdaIntegration(getContactFunction)
    );

    // Update Contact - PUT /contacts/{id}
    singleContactResource.addMethod(
      'PUT',
      new apigateway.LambdaIntegration(updateContactFunction)
    );

    // Delete Contact - DELETE /contacts/{id}
    singleContactResource.addMethod(
      'DELETE',
      new apigateway.LambdaIntegration(deleteContactFunction)
    ); */
  }

  private createLambdaFunction(
    id: string,
    entryPath: string,
    table: dynamodb.Table
  ): lambdaNodejs.NodejsFunction {
    const createContactFunction = new lambdaNodejs.NodejsFunction(this, id, {
      functionName: 'CreateContactFunction',
      runtime: lambda.Runtime.NODEJS_20_X,
      timeout: cdk.Duration.seconds(10),
      entry: path.join(__dirname, entryPath),
      handler: 'handler',
      environment: {
        TABLE_NAME: table.tableName,
      },
    });

    table.grantWriteData(createContactFunction);

    return createContactFunction;
  }

  private getLambdaFunction(
    id: string,
    entryPath: string,
    table: dynamodb.Table
  ): lambdaNodejs.NodejsFunction {
    const getContactFunction = new lambdaNodejs.NodejsFunction(this, id, {
      runtime: lambda.Runtime.NODEJS_20_X,
      entry: path.join(__dirname, entryPath),
      handler: 'handler',
      environment: {
        TABLE_NAME: table.tableName,
      },
    });
    table.grantReadData(getContactFunction);

    return getContactFunction;
  }
}
